import json
from typing import Optional, List
import requests

from e2enetworks.cloud.tir.skus import Plans, client
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.cloud.tir.helpers import plan_name_to_sku_item_price_id, plan_name_to_sku_id
from e2enetworks.constants import (
    CLIENT_NOT_READY_MESSAGE, NOTEBOOK,
    CUSTOM, headers,
)
from e2enetworks.cloud.tir.notebooks.constants import (
    COMMITTED_POLICY_REQUIRED_ERROR,
    INVALID_COMMITTED_POLICY_ERROR,
    NEXT_SKU_ITEM_PRICE_ID_REQUIRED_ERROR,
    IMAGE_VERSION_ID_REQUIRED_ERROR,
    CUSTOM_IMAGE_DETAILS_REQUIRED_ERROR,
    NOTEBOOK_ID_TYPE_ERROR, AUTO_RENEW_STATUS,
    AUTO_TERMINATE_STATUS, CONVERT_TO_HOURLY_BILLING,
    INVALID_NOTEBOOK_ID_TYPE_ERROR,
    INVALID_PLAN_NAME_TYPE_ERROR,
    INVALID_SKU_ITEM_PRICE_ID_TYPE_ERROR,
    INVALID_COMMITTED_POLICY_TYPE_ERROR,
    COMMITTED_POLICIES, SSH_ENABLE,
    SSH_DISABLE, SSH_UPDATE, PAID_USAGE,
    INSTANCE_TYPE, AVAILABLE_ACTIONS,
    INVALID_ACTION, UPDATE_ACTION_ERROR,
    ENABLE_ACTION_ERROR,
)


class Notebooks:
    def __init__(
            self,
            team: Optional[str] = "",
            project: Optional[str] = ""
    ):
        if not client.Default.ready():
            raise ValueError(CLIENT_NOT_READY_MESSAGE)
        if project:
            client.Default.set_project(project)
        if team:
            client.Default.set_team(team)

    def create(
        self,
        name: str,
        plan_name: str,
        image_type: str,
        image_version_id: Optional[int] = None,
        dataset_id_list: Optional[List[int]] = None,
        disk_size_in_gb: int = 30,
        is_jupyterlab_enabled: bool = True,
        public_key: Optional[List[str]] = None,
        notebook_type: str = "new",
        notebook_url: str = "",
        registry_namespace_id: Optional[int] = None,
        e2e_registry_image_url: Optional[str] = None,
        sku_item_price_id: Optional[int] = None,
        commited_policy: Optional[str] = None,
        next_sku_item_price_id: Optional[int] = None,
    ):
        # Avoid mutable defaults
        dataset_id_list = dataset_id_list or []
        public_key = public_key or []

        # Fetch SKU items based on the plan
        skus, skus_table = Plans().get_plans_list(NOTEBOOK, image_version_id)
        hourly_sku_item_price_id = plan_name_to_sku_item_price_id(skus, plan_name)
        sku_item_price_id = sku_item_price_id or hourly_sku_item_price_id

        # Validation for commited_policy and related fields
        if sku_item_price_id != hourly_sku_item_price_id and not commited_policy:
            raise ValueError(COMMITTED_POLICY_REQUIRED_ERROR)

        if commited_policy and commited_policy not in COMMITTED_POLICIES:
            raise ValueError(INVALID_COMMITTED_POLICY_ERROR.format(COMMITTED_POLICIES))

        if commited_policy in [AUTO_RENEW_STATUS, CONVERT_TO_HOURLY_BILLING] and not next_sku_item_price_id:
            raise ValueError(NEXT_SKU_ITEM_PRICE_ID_REQUIRED_ERROR)

        if commited_policy == AUTO_TERMINATE_STATUS:
            next_sku_item_price_id = None
        elif commited_policy == CONVERT_TO_HOURLY_BILLING:
            next_sku_item_price_id = hourly_sku_item_price_id

        # Validate image parameters
        if image_type != CUSTOM and not image_version_id:
            raise ValueError(IMAGE_VERSION_ID_REQUIRED_ERROR)

        if image_type == CUSTOM and (not registry_namespace_id or not e2e_registry_image_url):
            raise ValueError(CUSTOM_IMAGE_DETAILS_REQUIRED_ERROR)

        payload = {
            "name": name,
            "dataset_id_list": dataset_id_list,
            "image_type": image_type,
            "is_jupyterlab_enabled": is_jupyterlab_enabled,
            "public_key": public_key,
            "sku_item_price_id": sku_item_price_id,
            "auto_shutdown_timeout": None,
            "instance_type": PAID_USAGE,
            "disk_size_in_gb": disk_size_in_gb,
            "notebook_type": notebook_type,
            "notebook_url": notebook_url,
        }

        if image_type == CUSTOM:
            payload["registry_namespace_id"] = registry_namespace_id
            payload["e2e_registry_image_url"] = e2e_registry_image_url
        else:
            payload["image_version_id"] = image_version_id

        if commited_policy:
            payload["committed_instance_policy"] = commited_policy
            payload["next_sku_item_price_id"] = next_sku_item_price_id

        url = f"{client.Default.gpu_team_projects_path()}/notebooks/?"
        req = requests.Request('POST', url, json=payload)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def get(self, notebook_id: int):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def list(self):
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def delete(self, notebook_id: int):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/?"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def start(self, notebook_id: int):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/actions/?action=start&"
        req = requests.Request('PUT', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def stop(self, notebook_id: int):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/actions/?action=stop&"
        req = requests.Request('PUT', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def configure_ssh(self, notebook_id: int, action: str, ssh_keys_to_add=[], ssh_keys_to_remove=[]):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        if isinstance(ssh_keys_to_add, str):
            ssh_keys_to_add = [ssh_keys_to_add]
        if isinstance(ssh_keys_to_remove, str):
            ssh_keys_to_remove = [ssh_keys_to_remove]

        if action not in AVAILABLE_ACTIONS:
            raise ValueError(INVALID_ACTION.format(actions=AVAILABLE_ACTIONS))
        if action == SSH_UPDATE and not ssh_keys_to_add and not ssh_keys_to_remove:
            raise ValueError(UPDATE_ACTION_ERROR)
        if action == SSH_ENABLE and not ssh_keys_to_add:
            raise ValueError(ENABLE_ACTION_ERROR)

        if action == SSH_DISABLE:
            disable_url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/ssh-keys/"
            req = requests.Request('PUT', disable_url, json={"ssh_keys_to_remove": "all"})
        elif action == SSH_ENABLE:
            enable_url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/ssh-keys/"
            req = requests.Request('PUT', enable_url, json={"ssh_keys_to_add": ssh_keys_to_add})
        elif action == SSH_UPDATE:
            update_url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/ssh-keys/"
            req = requests.Request('PUT', update_url, json={"ssh_keys_to_add": ssh_keys_to_add, "ssh_keys_to_remove": ssh_keys_to_remove})
        else:
            raise ValueError(INVALID_ACTION)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def list_attached_ssh_keys(self, notebook_id: int):
        if not isinstance(notebook_id, int):
            raise TypeError(NOTEBOOK_ID_TYPE_ERROR)
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/ssh-keys/"
        req = requests.Request('GET', url)
        return prepare_object(client.Default.make_request(req))

    def upgrade(
        self,
        notebook_id: int,
        plan_name: str,
        sku_item_price_id: int,
        commited_policy: str,
        next_sku_item_price_id: Optional[int] = None
    ):
        if not isinstance(notebook_id, int):
            raise TypeError(INVALID_NOTEBOOK_ID_TYPE_ERROR)

        if not isinstance(plan_name, str):
            raise TypeError(INVALID_PLAN_NAME_TYPE_ERROR)

        if not isinstance(sku_item_price_id, (int, type(None))):
            raise TypeError(INVALID_SKU_ITEM_PRICE_ID_TYPE_ERROR)

        if not isinstance(commited_policy, str):
            raise TypeError(INVALID_COMMITTED_POLICY_TYPE_ERROR)

        skus, skus_table = Plans().get_plans_list(NOTEBOOK)
        hourly_sku_item_price_id = plan_name_to_sku_item_price_id(skus, plan_name)
        sku_id = plan_name_to_sku_id(skus, plan_name)
        sku_item_price_id = sku_item_price_id if sku_item_price_id else hourly_sku_item_price_id

        if sku_item_price_id != hourly_sku_item_price_id and not commited_policy:
            raise ValueError(COMMITTED_POLICY_REQUIRED_ERROR)

        if commited_policy not in COMMITTED_POLICIES:
            raise ValueError(INVALID_COMMITTED_POLICY_ERROR.format(COMMITTED_POLICIES))

        if commited_policy in [AUTO_RENEW_STATUS, CONVERT_TO_HOURLY_BILLING] and not next_sku_item_price_id:
            raise ValueError(NEXT_SKU_ITEM_PRICE_ID_REQUIRED_ERROR)

        if commited_policy == AUTO_TERMINATE_STATUS:
            next_sku_item_price_id = None
        elif commited_policy == CONVERT_TO_HOURLY_BILLING:
            next_sku_item_price_id = hourly_sku_item_price_id

        payload = {
            "sku_id": sku_id,
            "sku_item_price_id": sku_item_price_id,
            "committed_instance_policy": commited_policy,
            "next_sku_item_price_id": next_sku_item_price_id
        }
        headers['Authorization'] = f'Bearer {client.Default.access_token()}'
        url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/?apikey={client.Default.api_key()}"
        response = requests.put(url=url, headers=headers, json=payload)
        return prepare_object(response)

    def upgrade_pvc(self, notebook_id: int, size: int):
        if not isinstance(notebook_id, int):
            raise TypeError(INVALID_NOTEBOOK_ID_TYPE_ERROR)
        if not isinstance(size, int):
            raise TypeError(size)

        payload = json.dumps({
            "size": size})
        headers['Authorization'] = f'Bearer {client.Default.access_token()}'
        url = url = f"{client.Default.gpu_team_projects_path()}/notebooks/{notebook_id}/pvc/upgrade/?apikey={client.Default.api_key()}"
        response = requests.put(url=url, headers=headers, data=payload)
        return prepare_object(response)

    def registry_namespace_list(self):
        url = f"{client.Default.gpu_team_projects_path()}/container_registry/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def registry_detail(self, registry_namespace_id: int):
        url = f"{client.Default.gpu_team_projects_path()}/container_registry/{registry_namespace_id}/namespace-repository/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def commited_policy(self):
        print(COMMITTED_POLICIES)

    @staticmethod
    def help():
        help_text = """
    "Notebook Class Help"
    "================="
    The `Notebooks` class provides methods to interact with notebooks in a project.

    Available methods:
    1. create(
        name (required): str,
        plan_name (required): str - The name of the plan. You can list plans using the Plans API,
        image_type (required): str - The type of image ('custom' or predefined),
        image_version_id (optional): int - The version ID of the image,
        dataset_id_list (optional): list[int] - List of dataset IDs to attach to the notebook,
        disk_size_in_gb: int - Size of the disk in GB (default: 30),
        is_jupyterlab_enabled: bool - Enable/disable JupyterLab (default: True),
        public_key (optional): list[str] - List of SSH public keys,
        notebook_type: str - The type of notebook ('new' or 'existing') (default: 'new'),
        notebook_url: str - URL of an existing notebook (if notebook_type is 'existing'),
        registry_namespace_id (optional): int - Namespace ID for custom images,
        e2e_registry_image_url (optional): str - URL for the custom image in the registry,
        sku_item_price_id (optional): int - SKU item price ID for the notebook,
        commited_policy (optional): str - Committed policy type (required if SKU differs from hourly pricing),
        next_sku_item_price_id (optional): int - SKU item price ID for the next committed instance
    ): Create a new notebook with the specified parameters.

    2. get(notebook_id: int): Retrieve details about a specific notebook by its ID.

    3. list(): List all notebooks available in the current project.

    4. delete(notebook_id: int): Delete a notebook by its ID.

    5. start(notebook_id: int): Start a stopped notebook by its ID.

    6. stop(notebook_id: int): Stop a running notebook by its ID.

    7. configure_ssh(
        notebook_id: int,
        action: str - Action to perform ('enable', 'disable', 'update'),
        ssh_keys_to_add (optional): list[str] - List of SSH keys to add,
        ssh_keys_to_remove (optional): list[str] - List of SSH keys to remove
    ): Manage SSH keys for a notebook by enabling, disabling, or updating them.

    8. list_attached_ssh_keys(notebook_id: int): List all SSH keys attached to a notebook.

    9. upgrade(
        notebook_id: int,
        plan_name: str - The name of the new plan for upgrading,
        sku_item_price_id: int - The SKU item price ID for the upgrade,
        commited_policy: str - The committed policy type for the upgrade,
        next_sku_item_price_id (optional): int - The SKU item price ID for the next committed instance
    ): Upgrade the plan of an existing notebook.

    10. upgrade_pvc(notebook_id: int, size: int): Upgrade the PVC (Persistent Volume Claim) for a notebook.

    11. registry_namespace_list(): List all available container registry namespaces.

    12. registry_detail(registry_namespace_id: int): Retrieve details about a specific registry namespace by its ID.

    13. commited_policy(): Print all available committed policies.

    """
        print(help_text)



import urllib3

from e2enetworks.cloud.tir import version as tir_version

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
__version__ = tir_version.__version__

from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.apitoken import APITokens
from e2enetworks.cloud.tir.datasets.datasets import Datasets
from e2enetworks.cloud.tir.inference.endpoints import EndPoints
from e2enetworks.cloud.tir.images import Images
from e2enetworks.cloud.tir.inference.models import Models
from e2enetworks.cloud.tir.notebooks.notebook import Notebooks
from e2enetworks.cloud.tir.pipelines import PipelineClient
from e2enetworks.cloud.tir.projects import Projects
from e2enetworks.cloud.tir.skus import Plans
from e2enetworks.cloud.tir.teams import Teams
from e2enetworks.cloud.tir.api_client import ModelAPIClient
from e2enetworks.cloud.tir.finetuning.finetuner import FinetuningClient
from e2enetworks.cloud.tir.vector_db.vector_db import VectorDB
from e2enetworks.cloud.tir.distributed_jobs.distributed_job import DistributedJobClient
from e2enetworks.cloud.tir.integrations.integrations import Integration


init = client.Default.init
load_config = client.Default.load_config

__all__ = (
    "init",
    "PipelineClient",
    "load_config"
)


def help():
    """
AI Platform Help:

Provides an overview of available classes and methods in the platform.

Available classes:

- init: Provides functionalities for initialization.
- Images: Provides functionalities to interact with images.
- Plans: Provides functionalities to interact with skus.
- Integration: Provides functionalities to interact with integrations.
- Teams: Provides functionalities to interact with teams.
- Projects: Provides functionalities to interact with projects.
- Notebooks: Provides functionalities to interact with notebooks.
- Datasets: Provides functionalities to interact with datasets.
- EndPoints: Provides functionalities to interact with endpoints.
- Models: Provides functionalities to interact with models.
- PipelineClient: Provides functionalities to interact with Pipelines.
- APITokens: Provides functionalities to interact with API tokens.
- VectorDB: Provides functionalities to interact with Vector Databases.
    """

    # Call help() on each class
    class_objects = [
        client.Default, Images, Plans, Teams, Projects, Notebooks,
        Datasets, EndPoints, PipelineClient, Models, APITokens,
        ModelAPIClient, FinetuningClient, VectorDB, DistributedJobClient,
        Integration,
    ]

    for obj in class_objects:
        obj.help()



import base64


def convert_to_base64(text_to_encode):
    encoded_text = base64.b64encode(text_to_encode.encode()).decode()
    return encoded_text


def decode_base64(encoded_string):
    decoded_bytes = base64.b64decode(encoded_string)
    decoded_text = decoded_bytes.decode('utf-8')
    return decoded_text



from typing import Optional
import requests
from prettytable import PrettyTable

from e2enetworks.cloud.tir.finetuning.constants import (
    TEXT_MODELS_LIST, PLAN_NAME_ERROR,
    DEFAULT_TEXT_TRAINING_ARGS, DATASET_TYPES_LIST,
    DATASET_TYPE_ERROR, DEFAULT_IMAGE_TRAINING_ARGS,
    IMAGE_MODELS_LIST, INVALID_MODEL_NAME,
    SUPPORTED_MODELS_NOT_FOUND, FINETUNED,
    FAILED_TO_SHOW_FINETUNINGS, POD_NOT_FOUND,
    DETAILED_INFO, DEFAULT_FINETUNING_MODEL,
    SUCCEEDED, FINETUNING_NOT_SUCCEEDED,
    FINETUNING_MODEL_PATH, MODEL_ID_NOT_FOUND,
    )
from e2enetworks.constants import (
    INFERENCE, PIPELINE, CLIENT_NOT_READY_MESSAGE
    )
from e2enetworks.cloud.tir.skus import Plans, client
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.cloud.tir.finetuning.helpers import (
    get_supported_models_table, get_finetunings_table,
    get_sku_item_price_from_plan_name,
    get_model_training_inputs, get_finetuning_details,
    prepare_custom_endpoint_details,
    )


class FinetuningClient:
    def __init__(self, project: Optional[str] = None):
        if not client.Default.ready():
            raise ValueError(CLIENT_NOT_READY_MESSAGE)
        if project:
            client.Default.set_project(project)

    def create_finetuning(
        self,
        name: str,
        model_name: str,
        plan_name: str,
        huggingface_integration_id: str | int,
        dataset: str,
        dataset_type: str = "huggingface",
        wandb_integration_id: int = None,
        wandb_integration_run_name: str = "",
        description: str = None,
        training_type: str = "Peft",
        **training_args: dict
    ):
        if not isinstance(plan_name, str):
            return ValueError(PLAN_NAME_ERROR)
        if dataset_type not in DATASET_TYPES_LIST:
            raise ValueError(DATASET_TYPE_ERROR)

        if model_name in TEXT_MODELS_LIST:
            training_inputs = get_model_training_inputs(
                dataset,
                dataset_type,
                DEFAULT_TEXT_TRAINING_ARGS,
                **training_args)
        elif model_name in IMAGE_MODELS_LIST:
            training_inputs = get_model_training_inputs(
                dataset,
                dataset_type,
                DEFAULT_IMAGE_TRAINING_ARGS,
                **training_args)
        else:
            raise Exception(INVALID_MODEL_NAME)

        payload = {
            "name": name,
            "model_name": model_name,
            "huggingface_integration_id": huggingface_integration_id,
            "sku_item_price_id": get_sku_item_price_from_plan_name(plan_name=plan_name, service=PIPELINE, framework=None),
            "training_inputs": training_inputs,
            "training_type": training_type,
            "wandb_integration_id": wandb_integration_id if wandb_integration_id else None,
            "wandb_integration_run_name": wandb_integration_run_name,
            "description": description,
        }
        url = f"{client.Default.gpu_projects_path()}/finetuning/?"
        req = requests.Request('POST', url, json=payload)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def delete_finetuning(self, finetuning_id: str | int):
        url = f"{client.Default.gpu_projects_path()}/finetuning/{finetuning_id}/?"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def stop_finetuning(self, finetuning_id: str | int):
        url = f"{client.Default.gpu_projects_path()}/finetuning/{finetuning_id}/?&action=terminate&"
        req = requests.Request('PUT', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def retry_finetuning(self, finetuning_id: str | int):
        url = f"{client.Default.gpu_projects_path()}/finetuning/{finetuning_id}/?&action=retry&"
        req = requests.Request('PUT', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def create_finetuning_inference(
        self,
        inference_name: str,
        finetuning_id: str | int,
        huggingface_integration_id: int,
        plan_name: str,
        disk_size: int = 120,
    ):
        is_success, data = self.get_finetuning_details(finetuning_id)
        if not is_success:
            raise Exception(FAILED_TO_SHOW_FINETUNINGS)
        if data.status != SUCCEEDED:
            raise Exception(FINETUNING_NOT_SUCCEEDED.format(status=data.status))
        if not data.model_details:
            raise Exception(MODEL_ID_NOT_FOUND.format(finetuning_id=finetuning_id))

        DETAILED_INFO["hugging_face_id"] = data.model_name if data.model_name else DEFAULT_FINETUNING_MODEL
        custom_endpoint_details = prepare_custom_endpoint_details(data, disk_size)

        payload = {
            "name": inference_name,
            "custom_endpoint_details": custom_endpoint_details,
            "model_id": data.model_details.id,
            "sku_item_price_id": get_sku_item_price_from_plan_name(plan_name, INFERENCE, FINETUNED),
            "committed_instance_policy": "",
            "replica": 1,
            "committed_replicas": 0,
            "path": FINETUNING_MODEL_PATH,
            "framework": FINETUNED,
            "is_auto_scale_enabled": False,
            "detailed_info": DETAILED_INFO,
            "model_load_integration_id": huggingface_integration_id,
        }
        url = f"{client.Default.gpu_team_projects_path()}/serving/inference/?"
        req = requests.Request('POST', url, json=payload)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def show_all_finetunings(self):
        is_success, data = self.get_all_finetunings()
        if not is_success:
            print(FAILED_TO_SHOW_FINETUNINGS)
            return
        finetunings_table = get_finetunings_table(data)
        print(finetunings_table)

    def show_finetuning_details(self, finetuning_id: str | int):
        is_success, data = self.get_finetuning_details(finetuning_id)
        if not is_success:
            print(FAILED_TO_SHOW_FINETUNINGS)
            return
        finetuning_details = get_finetuning_details(data)
        print(finetuning_details)

    def show_finetuning_logs(self, finetuning_id: str | int):
        is_success, data = self.get_finetuning_logs(finetuning_id)
        if not is_success:
            print(FAILED_TO_SHOW_FINETUNINGS)
            return
        print(data)

    def show_supported_models(self):
        is_success, data = self.get_supported_models()
        if not is_success:
            print(SUPPORTED_MODELS_NOT_FOUND)
            return
        supported_models_table = get_supported_models_table(data)
        print(supported_models_table)

    def show_plan_names(self):
        plans = Plans()
        plans_list = plans.list(PIPELINE)
        gpu_skus = plans_list["GPU"]
        plans_table = PrettyTable()
        plans_table.field_names = ['name', 'series', 'cpu', 'gpu', 'memory',
                                   'sku_item_price_id', 'sku_type', 'committed_days', 'unit_price']
        plans.insert_plans_in_table(gpu_skus, plans_table)
        print(plans_table)

    def show_text_model_training_inputs(self):
        """
        Prints the training inputs for text models with their default values.
        """
        print("Text Model Training Inputs:")
        for key, (type_, default, _) in DEFAULT_TEXT_TRAINING_ARGS.items():
            print(f"- {key}: {type_.__name__} (default: {default})")

    def show_image_model_training_inputs(self):
        """
        Prints the training inputs for image models with their default values.
        """
        print("Image Model Training Inputs:")
        for key, (type_, default, _) in DEFAULT_IMAGE_TRAINING_ARGS.items():
            print(f"- {key}: {type_.__name__} (default: {default})")

    def get_all_finetunings(self):
        url = f"{client.Default.gpu_projects_path()}/finetuning/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def get_finetuning_details(self, finetuning_id: str | int):
        url = f"{client.Default.gpu_projects_path()}/finetuning/{finetuning_id}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def get_finetuning_logs(self, finetuning_id: str | int):
        is_success, data = self.get_finetuning_details(finetuning_id)
        if not is_success:
            raise Exception(FAILED_TO_SHOW_FINETUNINGS)
        run_pods = data.run_pods
        if not run_pods:
            raise Exception(POD_NOT_FOUND.format(finetuning_id=finetuning_id))
        url = f"{client.Default.gpu_projects_path()}/finetuning/{finetuning_id}/logs/?podname={run_pods[0].id}&"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def get_supported_models(self):
        url = f"{client.Default.gpu_projects_path()}/finetuning/model_types/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    @staticmethod
    def help():
        help_text = """
        FinetuningClient Class Help:

        This class provides methods for interacting with finetuning-related operations.
        Before using these methods, make sure to initialize the client using:
        - Using e2enetworks.cloud.tir.init(...)

        Available Methods:
        1. create_finetuning(
            self,
            name,
            model_name,
            plan_name,
            huggingface_integration_id,
            dataset,
            dataset_type,
            wandb_integration_id=None,
            wandb_integration_run_name="",
            description=None,
            training_type="Peft",
            **training_args
        )
            - Create a new finetuning.
            Note: Training inputs should be passed in **training_args as a dictionary.

        2. delete_finetuning(finetuning_id)
            - Delete a specific finetuning.

        3. stop_finetuning(finetuning_id)
            - Stop a specific finetuning.

        4. retry_finetuning(finetuning_id)
            - Retry a failed/terminated finetuning.

        5. create_finetuning_inference(
            self,
            inference_name,
            finetuning_id,
            huggingface_integration_id,
            plan_name,
            disk_size=120
        )
            - Create a new inference endpoint from a finetuning.

        6. show_all_finetunings()
            - Display a table of all finetunings.

        7. show_finetuning_details(finetuning_id)
            - Display detailed information about a specific finetuning.

        8. show_finetuning_logs(finetuning_id)
            - Show the finetuning logs.

        9. show_supported_models()
            - List currently supported models for finetuning.

        10. show_plan_names()
            - List currently supported SKUs for finetuning.

        11. show_text_model_training_inputs()
            - Print text model training inputs and defaults.

        12. show_image_model_training_inputs()
            - Print image model training inputs and defaults.

        13. get_all_finetunings()
            - Get details of all existing finetunings.

        14. get_finetuning_details(finetuning_id)
            - Get details of a specific finetuning.

        15. get_finetuning_logs(finetuning_id)
            - Get details of finetuning logs.

        16. get_supported_models()
            - Get currently supported models for finetuning.

        Note: Certain methods require specific arguments. Refer to the method signatures for details.
        """
        print(help_text)



from typing import Optional
import requests
from e2enetworks.cloud.tir import client
from e2enetworks.constants import CLIENT_NOT_READY_MESSAGE
from e2enetworks.cloud.tir.integrations.constants import (
    INTEGRATION_TYPES, INTEGRATION_TYPE_ERROR,
    WEIGHTS_BIASES, WANDB_USERNAME_ERROR,
    WANDB_PROJECT_ERROR, INTEGRATION_LIST_ERROR,
    SHOW_DETAILS_ERROR,
)
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.cloud.tir.integrations.helpers import (
    prepare_payload, get_integrations_table,
    get_details_table,
)


class Integration:
    def __init__(
        self,
        team: Optional[str] = "",
        project: Optional[str] = ""
    ):
        if not client.Default.ready():
            raise ValueError(CLIENT_NOT_READY_MESSAGE)
        if project:
            client.Default.set_project(project)
        if team:
            client.Default.set_team(team)

    def create_integration(
        self,
        name: str,
        token: str,
        integration_type: str,
        wandb_username: Optional[str] = None,
        wandb_project: Optional[str] = None,
        _get_raw: Optional[bool] = False
    ):
        if integration_type not in INTEGRATION_TYPES:
            raise ValueError(INTEGRATION_TYPE_ERROR.format(
                given_type=integration_type, integration_types=INTEGRATION_TYPES
            ))

        if integration_type == WEIGHTS_BIASES:
            if not wandb_username:
                raise ValueError(WANDB_USERNAME_ERROR)
            if not wandb_project:
                raise ValueError(WANDB_PROJECT_ERROR)
            payload = prepare_payload(
                name, token, integration_type, wandb_username=wandb_username, wandb_project=wandb_project
            )
        else:
            payload = prepare_payload(name, token, integration_type)

        url = f"{client.Default.gpu_team_projects_path()}/integrations/?"
        req = requests.Request('POST', url, json=payload)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def list_integrations(
        self,
        integration_type: str,
        _get_raw: Optional[bool] = False
    ):
        if integration_type not in INTEGRATION_TYPES:
            raise ValueError(INTEGRATION_TYPE_ERROR.format(
                given_type=integration_type, integration_types=INTEGRATION_TYPES
            ))

        url = f"{client.Default.gpu_team_projects_path()}/integrations/?integration_type={integration_type}&"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def get_integration_details(
        self,
        integration_id: int,
        _get_raw: Optional[bool] = False
    ):
        url = f"{client.Default.gpu_team_projects_path()}/integrations/{integration_id}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def delete_integration(
        self,
        integration_id: int,
        _get_raw: Optional[bool] = False
    ):
        url = f"{client.Default.gpu_team_projects_path()}/integrations/{integration_id}/?"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def show_integrations(
        self,
        integration_type: str,
    ):
        is_success, response = self.list_integrations(integration_type)
        if not is_success:
            print(INTEGRATION_LIST_ERROR)
            return
        integrations_table = get_integrations_table(response)

        if not integrations_table:
            print(INTEGRATION_LIST_ERROR)
            return
        print(integrations_table)

    def show_integration_details(
        self,
        integration_id: int
    ):
        is_success, response = self.get_integration_details(integration_id)
        if not is_success:
            print(SHOW_DETAILS_ERROR)
            return

        integration_details_table = get_details_table(response)
        if not integration_details_table:
            print(SHOW_DETAILS_ERROR)
            return
        print(integration_details_table)

    @staticmethod
    def help():
        help_text = f"""
        Integration Class Help:

        Available Methods:

        1. create_integration(name, token, integration_type={INTEGRATION_TYPES}, wandb_username=None, wandb_project=None)
           - Creates a new integration (Hugging Face or Weights & Biases).

        2. list_integrations(integration_type)
           - Lists all integrations of a specified type.

        3. get_integration_details(integration_id)
           - Retrieves details for a specific integration.

        4. delete_integration(integration_id)
           - Deletes a specific integration.

        5. show_integrations(integration_type)
           - Displays a table of all integrations.

        6. show_integration_details(integration_id)
           - Displays a table of detailed information about a specific integration.
        """
        print(help_text)



import requests

from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.constants import BASE_GPU_URL


class SSHKeys:

    def __init__(self):
        client_not_ready = (
            "Client is not ready. Please initiate client by:"
            "\n- Using e2enetworks.cloud.tir.init(...)"
        )
        if not client.Default.ready():
            raise ValueError(client_not_ready)

    def add(self, label, ssh_key):
        url = f"{BASE_GPU_URL}ssh_keys/"
        req = requests.Request('POST', url, json={"label": label, "ssh_key": ssh_key})
        response = client.Default.make_request(req)
        if response.status_code == 200:
            print("Successfully added SSH key.")
        return prepare_object(response)

    def list(self):
        url = f"{BASE_GPU_URL}ssh_keys/"
        req = requests.Request('GET', url)
        return prepare_object(client.Default.make_request(req))

    def sync(self):
        url = f"{BASE_GPU_URL}ssh_keys/sync/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        if response.status_code == 200:
            print("Successfully synced.")
        return prepare_object(response)

    def delete(self, id):
        url = f"{BASE_GPU_URL}ssh_keys/{id}/"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        if response.status_code == 200:
            print("Successfully deleted SSH key.")
        return prepare_object(response)

    @staticmethod
    def help():
        print("\t\tSSHKeys Class Help")
        print("\t\t===================")
        help_text = """
                SSHKeys class provides methods for ssh-keys related operations.

                Available methods:
                1. Add a new ssh-key.
                    add(
                        label(required): String,
                        ssh_key(required): String
                    )
                2. List all ssh-keys.
                    list()
                3. Sync ssh-keys with my-account.
                    sync()
                4. Delete a ssh-key.
                    delete(
                        id(required): integer => obtain from list() method
                    )

                Usage:
                ssh_key = SSHKeys()
                ssh_key.add(label, ssh_key)
                ssh_key.list()
                ssh_key.sync()
                ssh_key.delete(id)
                """
        print(help_text)



from prettytable import PrettyTable
from typing import Type
from e2enetworks.cloud.tir.finetuning.constants import (
    HUGGING_FACE, INVALID_DATASET,
    FINETUNING_HF_MODEL_ID_TO_IMAGE_MAPPING,
    CUSTOM_ENDPOINT_DETAILS, PLAN_NAME_INVALID,
    )
from e2enetworks.cloud.tir.skus import Plans


def get_supported_models_table(data):
    supported_models_table = PrettyTable()
    supported_models_table.field_names = ['Supported Fine-tuning Models']
    for model in data:
        supported_models_table.add_row([model.name])
    return supported_models_table


def get_finetunings_table(data):
    finetunings_table = PrettyTable()
    finetunings_table.field_names = ['Finetuning ID', 'Finetuning Name', 'SKU Name', 'Status']
    for item in data:
        finetunings_table.add_row([
            item.id,
            item.name,
            item.sku.name,
            item.status
        ])
    return finetunings_table


def get_sku_item_price_from_plan_name(plan_name, service, framework=None, committed_days=0):
    plans = Plans().list(service=service, framework=framework)
    plan_name_to_sku = {}
    gpu_skus = plans["GPU"]
    for sku in gpu_skus:
        for sku_item_price in sku["plans"]:
            if not sku["is_free"]:
                name = sku.get('name')
                committed_days = sku_item_price.get('committed_days')
                key = f'{name}_c{committed_days}'
                plan_name_to_sku[key] = sku_item_price['sku_item_price_id']
    if plan_name_to_sku.get(f'{plan_name}_c{committed_days}'):
        return plan_name_to_sku.get(f'{plan_name}_c{committed_days}')
    raise Exception(PLAN_NAME_INVALID.format(plan_name=plan_name))


def get_model_training_inputs(dataset, dataset_type, default_training_args, **training_args):
    return {
        **get_dataset_config(dataset, dataset_type),
        **{key: get_argument_from_training_args(key, training_args, type_, default, is_required)
            for key, (type_, default, is_required) in default_training_args.items()}
    }


def get_dataset_config(dataset, dataset_type):
    if dataset_type == HUGGING_FACE:
        return {"dataset_type": HUGGING_FACE,
                "dataset": dataset}
    dataset_sub_str = dataset.split('/')
    if len(dataset_sub_str) <= 1:
        raise Exception(INVALID_DATASET)
    object_name = dataset.replace(f"{dataset_sub_str[0]}/", '', 1)
    if not object_name:
        raise Exception("dataset invalid")
    return {"dataset_type": dataset_type,
            "dataset": f'{dataset_sub_str[0]}/{object_name}'}


def get_argument_from_training_args(argument_name, training_args, type_: Type, default="", is_required=False):
    """Get argument from training_args with a type check, default value, and required condition."""
    # Check if the argument is in training_args
    if argument_name in training_args:
        value = training_args[argument_name]
    elif is_required:
        raise ValueError(f"Required argument missing: {argument_name}")
    else:
        return default

    # Check type of the argument
    if not isinstance(value, type_):
        raise TypeError(f"Argument type is invalid: {argument_name}, valid type is {type_}.")

    return value


def get_finetuning_details(data):
    finished_at = data.run.finished_at if data.run else "-"
    finetuning_details = f"""
    Finetuning Details:
    -------------------
    Finetuning ID       : {data.id}
    Finetuning Name     : {data.name}
    Model Name          : {data.model_name}
    Dataset Name        : {data.training_inputs.dataset_name}
    Dataset Type        : {data.training_inputs.dataset_type}
    Training Type       : {data.training_type}
    Status              : {data.status}
    Created At          : {data.created_at}
    Finished At         : {finished_at}
    Huggingface ID      : {data.huggingface_integration_id}
    Description         : {data.description or "None"}
    SKU ID              : {data.sku.sku_id}
    SKU Name            : {data.sku.name}
    """
    return finetuning_details


def prepare_custom_endpoint_details(response, disk_size):
    image_name = FINETUNING_HF_MODEL_ID_TO_IMAGE_MAPPING.get(response.model_name)
    if image_name:
        CUSTOM_ENDPOINT_DETAILS["container"]["container_name"] = image_name
    CUSTOM_ENDPOINT_DETAILS["resource_details"]["disk_size"] = disk_size
    return CUSTOM_ENDPOINT_DETAILS



import yaml
import json
from typing import Dict, Any
from requests import Response
from types import SimpleNamespace


def load_yaml(
        path: str,
) -> Dict[str, Any]:
    _load_from_local(path)


def _load_from_local(path: str) -> Dict[str, Any]:
    with open(path) as f:
        return yaml.safe_load(f)


def load_json(b: bytes, object_hook=None) -> Dict:
    return json.loads(b, object_hook=object_hook)


def prepare_object(response: Response, _get_raw: bool = False):
    if not response.ok:
        try:
            output = response.json()['errors']
        except:
            output = response.reason
        print("ERROR:", output)
        return False, response

    if _get_raw:
        return True, response

    response = load_json(response.content, object_hook=lambda d: SimpleNamespace(**d))
    if hasattr(response, "data"):
        return True, response.data
    else:
        return True, response


def prepare_response(response: Response) -> Dict:
    if response.ok:
        return load_json(response.content)

    # logger.error("")
    print("api call failed with error:", response.reason)
    return {
        "error": response.reason
    }



# Error Messages
COMMITTED_POLICY_REQUIRED_ERROR = "Committed policy is required when SKU item price does not match hourly rate."
INVALID_COMMITTED_POLICY_ERROR = "Committed policy must be one of: {}"
NEXT_SKU_ITEM_PRICE_ID_REQUIRED_ERROR = "Next SKU item price ID is required for committed policy."
IMAGE_VERSION_ID_REQUIRED_ERROR = "Image version ID is required for non-custom images."
CUSTOM_IMAGE_DETAILS_REQUIRED_ERROR = "Both registry_namespace_id and e2e_registry_image_url are required for custom images."
NOTEBOOK_ID_TYPE_ERROR = "notebook_id must be a valid integer value."
INVALID_NOTEBOOK_ID_TYPE_ERROR = "Notebook ID must be an integer."
INVALID_PLAN_NAME_TYPE_ERROR = "Plan name must be a string."
INVALID_SKU_ITEM_PRICE_ID_TYPE_ERROR = "SKU item price ID must be an integer."
INVALID_COMMITTED_POLICY_TYPE_ERROR = "Committed policy must be a string."
INVALID_ACTION = "Invalid actions, must be one of {actions}."
UPDATE_ACTION_ERROR = "'update' action requires either 'ssh_keys_to_add' or 'ssh_keys_to_remove'"
ENABLE_ACTION_ERROR = "'enable' action requires 'ssh_keys_to_add'"

# Keywords
AUTO_RENEW_STATUS = 'auto_renew'
AUTO_TERMINATE_STATUS = 'auto_terminate'
CONVERT_TO_HOURLY_BILLING = 'convert_to_hourly_billing'
SSH_ENABLE = "enable"
SSH_DISABLE = "disable"
SSH_UPDATE = "update"
FREE_USAGE = "free_usage"
PAID_USAGE = "paid_usage"

# Lists and Dicts
COMMITTED_POLICIES = [AUTO_RENEW_STATUS, AUTO_TERMINATE_STATUS, CONVERT_TO_HOURLY_BILLING]
AVAILABLE_ACTIONS = [SSH_ENABLE, SSH_DISABLE, SSH_UPDATE]
INSTANCE_TYPE = [FREE_USAGE, PAID_USAGE]



from typing import Optional, Dict
from prettytable import PrettyTable
from e2enetworks.cloud.tir.integrations.constants import (
    WANDB_USERNAME, WANDB_KEY, INTEGRATION_TYPE,
    WANDB_PROJECT, INTEGRATION_DETAILS, NAME,
    HUGGINGFACE_TOKEN, WEIGHTS_BIASES,
)


def prepare_payload(
    name: str,
    token: str,
    integration_type: str,
    wandb_username: Optional[str] = None,
    wandb_project: Optional[str] = None
) -> Dict:
    """
    Prepares the payload for creating Hugging Face or Weights & Biases integration.

    Returns:
        Dict: A dictionary containing the prepared payload.
    """
    integration_details = {HUGGINGFACE_TOKEN: token}

    # Add additional details if it's Weights & Biases integration
    if integration_type == WEIGHTS_BIASES:
        integration_details = {
            WANDB_USERNAME: wandb_username,
            WANDB_KEY: token,
            WANDB_PROJECT: wandb_project
        }
    payload = {
        NAME: name,
        INTEGRATION_TYPE: integration_type,
        INTEGRATION_DETAILS: integration_details,
    }
    return payload


def get_integrations_table(response):
    """
    Prepares the table from the response of get_integrations function

    Returns:
        Dict: A table containing the details of all integrations
    """
    integrations_table = PrettyTable()
    integrations_table.field_names = ['ID', 'Name', 'Integration Type']
    for integration in response:
        integrations_table.add_row([
            integration.id,
            integration.name,
            integration.integration_type,
        ])
    return integrations_table


def get_details_table(response):
    """
    Prepares the table from the response of get_integration_details function

    Returns:
        Dict: A table containing the details of the integration id
    """
    details_table = PrettyTable()
    details_table.field_names = ['ID', 'Name', 'Integration Type', 'Created By', 'Created At', 'Integration Details']
    details_table.add_row([
        response.id,
        response.name,
        response.integration_type,
        response.created_by.email,
        response.created_at,
        response.integration_details
    ])
    return details_table



HUGGING_FACE = "huggingface"
EOS_BUCKET = "eos-bucket"
DATASET_TYPES_LIST = [HUGGING_FACE, EOS_BUCKET]

LLAMA2_7B_HF_MODEL_ID = 'meta-llama/Llama-2-7b-hf'
LLAMA3_8B_HF_MODEL_ID = 'meta-llama/Meta-Llama-3-8B'
LLAMA3_8B_INST_HF_MODEL_ID = 'meta-llama/Meta-Llama-3-8B-Instruct'
MISTRAL_7B_HF_MODEL_ID = 'mistralai/Mistral-7B-v0.1'
MISTRAL_7B_INST_HF_MODEL_ID = 'mistralai/Mistral-7B-Instruct-v0.2'
MISTRAL_8X7B_HF_MODEL_ID = 'mistralai/Mixtral-8x7B-v0.1'
GEMMA_7B_HF_MODEL_ID = 'google/gemma-7b'
GEMMA_7B_INST_HF_MODEL_ID = 'google/gemma-7b-it'
STABLE_DIFFUSION_2_HF_MODEL_ID = 'stabilityai/stable-diffusion-2-1'
STABLE_DIFFUSION_XL_HF_MODEL_ID = 'stabilityai/stable-diffusion-xl-base-1.0'

TEXT_MODELS_LIST = [
   LLAMA2_7B_HF_MODEL_ID, LLAMA3_8B_HF_MODEL_ID, LLAMA3_8B_INST_HF_MODEL_ID,
   MISTRAL_7B_HF_MODEL_ID, MISTRAL_7B_INST_HF_MODEL_ID, MISTRAL_8X7B_HF_MODEL_ID,
   GEMMA_7B_HF_MODEL_ID, GEMMA_7B_INST_HF_MODEL_ID,
   ]
IMAGE_MODELS_LIST = [
   STABLE_DIFFUSION_2_HF_MODEL_ID,
   STABLE_DIFFUSION_XL_HF_MODEL_ID,
   ]
COMBINED_MODELS_LIST = TEXT_MODELS_LIST + IMAGE_MODELS_LIST
PIPELINE = "pipeline"
INVALID_DATASET = "Invalid dataset"
PLAN_NAME_ERROR = "plan_name is should be a string"
DEFAULT_TEXT_TRAINING_ARGS = {
    "validation_split_ratio": (float, 0.1, False),
    "target_dataset_field": (str, 'text', False),
    "gradient_accumulation_steps": (int, 1, False),
    "context_length": (int, 512, False),
    "learning_rate": (float, 0.0000141, False),
    "epochs": (int, 3, False),
    "stop_training_when": (str, "epoch_count", False),
    "max_steps": (int, -1, False),
    "batch_size": (int, 4, False),
    "peft_lora_alpha": (int, 16, False),
    "peft_lora_r": (int, 64, False),
    "save_strategy": (str, "no", False),
    "task": (str, "Instruction-Finetuning", False),
    "prompt_configuration": (str, "", False),
    "save_steps": (int, 10, False),
    "limit_training_records_count": (int, -1, False),
    "limit_eval_records_count": (int, -1, False),
    "source_repository_type": (str, "base_model", False),
    "source_model_repo_info": (str, "", False),
    "source_model_path": (str, "", False),
    "save_total_limit": (int, 10, False),
    "lora_dropout": (float, 0.05, False),
    "lora_bias": (str, "none", False),
    "load_in_4bit": (bool, False, False),
    "bnb_4bit_compute_dtype": (str, "bfloat16", False),
    "bnb_4bit_quant_type": (str, "fp4", False),
    "bnb_4bit_use_double_quant": (bool, False, False),
}
DEFAULT_IMAGE_TRAINING_ARGS = {
    "gradient_accumulation_steps": (int, 1, False),
    "learning_rate": (float, 0.0000141, False),
    "epochs": (int, 3, False),
    "stop_training_when": (str, "epoch_count", False),
    "max_steps": (int, -1, False),
    "batch_size": (int, 4, False),
    "peft_lora_alpha": (int, 16, False),
    "peft_lora_r": (int, 64, False),
    "save_strategy": (str, "no", False),
    "task": (str, "Instruction-Finetuning", False),
    "save_steps": (int, 10, False),
    "limit_training_records_count": (int, -1, False),
    "source_repository_type": (str, "base_model", False),
    "source_model_repo_info": (str, "", False),
    "source_model_path": (str, "", False),
    "save_total_limit": (int, 10, False),
    "image_column": (str, "image", False),
    "caption_column": (str, "text", False),
    "validation_prompt": (str, "A photo of a man with green eyes", False),
    "num_validation_images": (int, 2, False),
    "train_batch_size": (int, 1, False),
    "load_in_8bit": (bool, False, False),
    "mixed_precision": (str, "no", False),
    "checkpointing_steps": (int, 500, False),
    "checkpoints_total_limit": (int, 10, False)
}
INVALID_MODEL_NAME = f"Model name must be in {COMBINED_MODELS_LIST}"
SUPPORTED_MODELS_NOT_FOUND = "Failed to fetch supported models"
FAILED_TO_SHOW_FINETUNINGS = "Failed to fetch finetuning details"
CUSTOM_ENDPOINT_DETAILS = {
    "service_port": False,
    "metric_port": False,
    "container": {
        "container_name": "vllm/vllm-openai:latest",
        "container_type": "public",
        "private_image_details": {},
        "advance_config": {
            "image_pull_policy": "Always",
            "is_readiness_probe_enabled": False,
            "is_liveness_probe_enabled": False,
            "readiness_probe": {
                "protocol": "http",
                "initial_delay_seconds": 10,
                "success_threshold": 1,
                "failure_threshold": 3,
                "port": 8080,
                "period_seconds": 10,
                "timeout_seconds": 10,
                "path": "/metrics",
                "grpc_service": "",
                "commands": ""
            },
            "liveness_probe": {
                "protocol": "http",
                "initial_delay_seconds": 10,
                "success_threshold": 1,
                "failure_threshold": 3,
                "port": 8080,
                "period_seconds": 10,
                "timeout_seconds": 10,
                "path": "/metrics",
                "grpc_service": "",
                "commands": ""
            }
        }
    },
    "resource_details": {
        "disk_size": 120,
        "mount_path": "",
        "env_variables": [
            {
                "key": "HF_HOME",
                "value": "/mnt/models",
                "required": True,
                "disabled": {
                    "key": True,
                    "value": True
                }
            }
        ]
    },
    "public_ip": "no"
}
DETAILED_INFO = {
    "commands": "",
    "args": "",
    "hugging_face_id": "",
    "tokenizer": "",
    "server_version": "",
    "world_size": 1,
    "error_log": True,
    "info_log": True,
    "warning_log": True,
    "log_verbose_level": 1,
    "model_serve_type": ""
}
FINETUNING_MODEL_PATH = "final/"
FINETUNED = "finetuned"

FINETUNED_VLLM_OPENAI_IMAGE = 'vllm/vllm-openai:latest'
FINETUNED_MISTRAL_8X7B_INFERENCE_IMAGE = 'registry.e2enetworks.net/aimle2e/finetuned-mixtral8x7b-server:v1'
FINETUNED_STABLE_DIFFUSION_2_INFERENCE_IMAGE = 'registry.e2enetworks.net/aimle2e/finetuned-stable-diff-2:v1'
FINETUNED_STABLE_DIFFUSION_XL_INFERENCE_IMAGE = 'registry.e2enetworks.net/aimle2e/finetuned-stable-diff-sdxl:v1'

FINETUNING_HF_MODEL_ID_TO_IMAGE_MAPPING = {
    MISTRAL_8X7B_HF_MODEL_ID: FINETUNED_MISTRAL_8X7B_INFERENCE_IMAGE,
    STABLE_DIFFUSION_2_HF_MODEL_ID: FINETUNED_STABLE_DIFFUSION_2_INFERENCE_IMAGE,
    STABLE_DIFFUSION_XL_HF_MODEL_ID: FINETUNED_STABLE_DIFFUSION_XL_INFERENCE_IMAGE,
    LLAMA2_7B_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    LLAMA3_8B_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    LLAMA3_8B_INST_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    MISTRAL_7B_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    MISTRAL_7B_INST_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    GEMMA_7B_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
    GEMMA_7B_INST_HF_MODEL_ID: FINETUNED_VLLM_OPENAI_IMAGE,
}
DEFAULT_FINETUNING_MODEL = LLAMA2_7B_HF_MODEL_ID
SUCCEEDED = 'Succeeded'
FINETUNING_NOT_SUCCEEDED = "Finetuning is not succeeded, it is in {status} state"
MODEL_ID_NOT_FOUND = "Model id not found for finetuning id {finetuning_id}"
PLAN_NAME_INVALID = "Plan name {plan_name} is invalid"
POD_NOT_FOUND = "No pods found for fine-tuning {finetuning_id}"
DATASET_TYPE_ERROR = "dataset_type must be either 'huggingface' or 'eos-bucket'"



import json

import requests
import os

from typing import Optional
from e2enetworks.constants import BASE_GPU_URL, BUCKET_TYPES, MODEL_TYPES, headers, MANAGED_STORAGE
from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.cloud.tir.minio_service import MinioService
from e2enetworks.cloud.tir.inference.constants import (
    INVALID_MODEL_ID, MODEL_LOAD_ERROR,
)
from transformers import AutoModel


class Models:
    def __init__(
            self,
            team: Optional[str] = "",
            project: Optional[str] = ""
    ):
        client_not_ready = (
            "Client is not ready. Please initiate client by:"
            "\n- Using e2enetworks.cloud.tir.init(...)"
        )
        if not client.Default.ready():
            raise ValueError(client_not_ready)

        if project:
            client.Default.set_project(project)

        if team:
            client.Default.set_team(team)

    def create(self, name, model_type='custom', storage_type=MANAGED_STORAGE, bucket_name=None, job_id=None, score={}, access_key=None, secret_key=None):
        try:
            payload = json.dumps({
                "name": name,
                "model_type": model_type,
                "bucket_name": bucket_name,
                "storage_type": storage_type,
                "finetuning_id": job_id,
                "score": score,
                "access_key": access_key,
                "secret_key": secret_key
            })
        except Exception as e:
            raise Exception(f"Input Error {e}")
        headers['Authorization'] = f'Bearer {client.Default.access_token()}'
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/?" \
              f"apikey={client.Default.api_key()}"
        response = requests.post(url=url, headers=headers, data=payload)
        response = prepare_object(response)
        return response

    def push_model(self, model_id, model_path, prefix="", model_type="custom", job_id=None, score={}, use_caching=False):
        is_success, model = self.get(model_id)
        if not is_success:
            raise ValueError(INVALID_MODEL_ID)

        access_key = model.access_key.access_key
        secret_key = model.access_key.secret_key
        minio_service = MinioService(access_key=access_key, secret_key=secret_key)
        if os.path.isdir(model_path):
            is_success, logs = minio_service.upload_directory_recursive(
                bucket_name=model.bucket.bucket_name,
                source_directory=model_path,
                prefix=prefix
            )
        else:
            is_success, logs = minio_service.upload_file(
                bucket_name=model.bucket.bucket_name,
                file_path=model_path,
                prefix=prefix
            )
        if use_caching:
            minio_service.update_cache_file(bucket_name=model.bucket.bucket_name)
        return is_success, logs

    def download_model(self, model_id, local_path, prefix=""):
        is_success, model = self.get(model_id)
        if not is_success:
            raise ValueError(INVALID_MODEL_ID)
        try:
            access_key = model.access_key.access_key
            secret_key = model.access_key.secret_key
            minio_service = MinioService(access_key=access_key, secret_key=secret_key)
            return minio_service.download_directory_recursive(
                bucket_name=model.bucket.bucket_name,
                local_path=local_path,
                prefix=prefix
            )
        except Exception as e:
            return False, str(e)

    def list_bucket_data(self, model_id, prefix=""):
        if not model_id:
            raise ValueError("model id is mandatory")
        model = self.get(model_id)
        if not model:
            raise ValueError("Invalid model id")

        headers['Authorization'] = f'Bearer {client.Default.access_token()}'
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/" \
              f"{model_id}/bucket/?apikey={client.Default.api_key()}&prefix={prefix}"
        response = requests.get(url=url, headers=headers)
        response = prepare_object(response)
        for i in response:
            print(i.name)
        return response

    def _update_repo(self, model_id, score={}):
        if not isinstance(model_id, int):
            raise ValueError(model_id)
        payload = {
            "score": score
        }

        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/{model_id}/"
        req = requests.Request('PUT', url, json=payload)
        response = client.Default.make_request(req)
        response = prepare_object(response)
        return response

    def get(self, model_id):
        if not isinstance(model_id, int):
            raise ValueError(model_id)

        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/" \
              f"{model_id}/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        response = prepare_object(response)
        return response

    def list(self):
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        response = prepare_object(response)
        return response

    def delete(self, model_id):
        if not isinstance(model_id, int):
            raise ValueError(model_id)

        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/serving/model/" \
              f"{model_id}/"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def load_model(
        self,
        model_id: int,
        local_path: Optional[str] = "",
        prefix: Optional[str] = ""
    ):
        """
        Downloads the model folder to a local directory and returns a HuggingFace model object.

        Args:
            model_id (int): The ID of the model repository.
            local_path (Optional[str]): The local directory to download the model to.
            prefix (Optional[str]): An optional path prefix where the model is located in the Model repo.

        Returns:
            model_object: A HuggingFace model object created using the AutoModel class.
        """
        try:
            # Download the model to the specified local path
            self.download_model(model_id, local_path, prefix)
            # Load the HuggingFace model from the downloaded directory
            model_object = AutoModel.from_pretrained(local_path)
            return model_object

        except Exception as e:
            error_message = MODEL_LOAD_ERROR.format(model_id=model_id, local_path=local_path, error_message=str(e))
            raise ValueError(error_message)

    @staticmethod
    def help():
        print("Models Class Help")
        print("\t\t=================")
        print("\t\tThis class provides functionalities to interact with models.")
        print("\t\tAvailable methods:")
        print(
            "\t\t1. __init__(team, project): Initializes a Models instance with the specified team and project "
            "IDs.")
        print("\t\t2. create(name, model_type, storage_type, bucket_name): Creates a new model with the provided "
              "details.")
        print("\t\t3. push_model(model_path, prefix, model_id, job_id, score, model_type, use_caching): Creates a new model with the provided "
              "details.")
        print("\t\t4. download_model(model_id, local_path, prefix)")
        print("\t\t5. list_bucket_data(model_id, prefix): to list the content of bucket")
        print("\t\t6. get(model_id): Retrieves information about a specific model using its ID.")
        print("\t\t7. list(): Lists all models associated with the team and project.")
        print("\t\t8. delete(model_id): Deletes a model with the given ID.")
        print("\t\t9. help(): Displays this help message.")

        # Example usages
        print("\t\tExample usages:")
        print("\t\tmodels = Models(123, 456)")
        print(f"\t\tmodels.create(name='Test Dataset', model_type={MODEL_TYPES}, , storage_type={BUCKET_TYPES}, "
              f"bucket_name='dataset-bucket'")
        print("\t\tmodels.push_model(model_path, prefix='', model_id=None, model_type='custom')")
        print("\t\tmodels.download_model(model_id=<model id>, local_path=<path of local directory>,"
              " prefix=<prefix in the bucket>)")
        print("\t\tmodels.list_bucket_data(model_id=<model id>, prefix='')")
        print("\t\tmodels.get(789)")
        print("\t\tmodels.list()")
        print("\t\tmodels.delete(789)")



# Error constants
ALLOWED_FILES_ERROR = "Given file format must be in {file_formats}"
DATASET_LOAD_ERROR = "Unable to load dataset: {error_message}"
INVALID_DATASET_ID_ERROR = "Invalid dataset Id: {dataset_id}"
DATASET_LISTING_ERROR = "Unable to list datasets"
NO_DATASETS_AVAILABLE = "No datasets available"
BUCKET_NAME_ERROR = "Please specify a valid bucket name"
DATASET_ID_ERROR = "Invalid dataset_id parameter"
BUCKET_TYPE_ERROR = "Bucket type must be in {bucket_types}"
UPLOAD_PATH_ERROR = "The provided path '{dataset_path}' does not exist."
UPLOAD_FILE_DIR_ERROR = "The provided path '{dataset_path}' is neither a file nor a directory."

# Dataset constraints
ALLOWED_FILE_FORMATS = ['csv', 'json', 'jsonl', 'txt', 'parquet']

# Storage
MANAGED_STORAGE = "managed"
E2E_OBJECT_STORAGE = "e2e_s3"
BUCKET_TYPES = [MANAGED_STORAGE, E2E_OBJECT_STORAGE]

# Encryption constants
E2E_MANAGED = "e2e-managed"
USER_MANAGED = "user-managed"
E2E_MANAGED_KEY = "sse-kms"
USER_MANAGED_KEY = "sse-c"
ENCRYPTION_TYPES = [E2E_MANAGED, USER_MANAGED]
ENCRYPTION_TYPES_MAPPING = {
    E2E_MANAGED: E2E_MANAGED_KEY,
    USER_MANAGED: USER_MANAGED_KEY,
}
ENCRYPTION_TYPE_ERROR = f"Encryption type must be in {ENCRYPTION_TYPES}"
ENCRYPTION_ENABLE_TYPE_ERROR = "encryption_enable must be of type 'bool'"



import json
from typing import Optional
from datasets import load_dataset

import requests
import os

from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.minio_service import MinioService
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.cloud.tir.datasets.helpers import (
    prepare_datasets_table
)
from e2enetworks.cloud.tir.helpers import get_file_extension
from e2enetworks.constants import (
    BASE_GPU_URL, headers,
)
from e2enetworks.cloud.tir.datasets.constants import (
    ALLOWED_FILE_FORMATS, ALLOWED_FILES_ERROR,
    DATASET_LOAD_ERROR, INVALID_DATASET_ID_ERROR,
    DATASET_LISTING_ERROR, NO_DATASETS_AVAILABLE,
    MANAGED_STORAGE, BUCKET_NAME_ERROR, ENCRYPTION_TYPES,
    BUCKET_TYPE_ERROR, ENCRYPTION_TYPE_ERROR,
    ENCRYPTION_TYPES_MAPPING, ENCRYPTION_ENABLE_TYPE_ERROR,
    BUCKET_TYPES, E2E_OBJECT_STORAGE, UPLOAD_FILE_DIR_ERROR,
    UPLOAD_PATH_ERROR,
)
from e2enetworks.constants import CLIENT_NOT_READY_MESSAGE


class Datasets:
    def __init__(
            self,
            team: Optional[str] = "",
            project: Optional[str] = "",
    ):
        if not client.Default.ready():
            raise ValueError(CLIENT_NOT_READY_MESSAGE)

        if project:
            client.Default.set_project(project)

        if team:
            client.Default.set_team(team)

    def create(
        self,
        name: str,
        bucket_name: Optional[str] = None,
        bucket_type: Optional[str] = MANAGED_STORAGE,
        encryption_enable: Optional[bool] = False,
        encryption_type: Optional[str] = None,
        description: Optional[str] = ""
    ):
        if bucket_type not in BUCKET_TYPES:
            raise ValueError(BUCKET_TYPE_ERROR.format(BUCKET_TYPES))
        if bucket_type == E2E_OBJECT_STORAGE:
            if not bucket_name:
                raise ValueError(BUCKET_NAME_ERROR)
        if not isinstance(encryption_enable, bool):
            raise TypeError(ENCRYPTION_ENABLE_TYPE_ERROR)
        if encryption_enable and encryption_type not in ENCRYPTION_TYPES:
            raise ValueError(ENCRYPTION_TYPE_ERROR)

        payload = json.dumps({
            "type": bucket_type,
            "name": name,
            "bucket_name": bucket_name,
            "encryption_enable": encryption_enable,
            "encryption_type": ENCRYPTION_TYPES_MAPPING.get(encryption_type),
            "description": description,
        })
        headers['Authorization'] = f'Bearer {client.Default.access_token()}'
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/datasets/?" \
              f"apikey={client.Default.api_key()}"
        response = requests.post(url=url, headers=headers, data=payload)
        return prepare_object(response)

    def get(self, dataset_id: int):
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/datasets/" \
              f"{dataset_id}/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def list(self):
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/datasets/"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def delete(self, dataset_id: int):
        url = f"{BASE_GPU_URL}teams/{client.Default.team()}/projects/{client.Default.project()}/datasets/" \
              f"{dataset_id}/"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response)

    def upload_dataset(
        self,
        dataset_id: int,
        upload_dataset_path: str,
        prefix: Optional[str] = ""
    ):
        if not os.path.exists(upload_dataset_path):
            raise FileNotFoundError(UPLOAD_PATH_ERROR.format(dataset_path=upload_dataset_path))
        if not os.path.isfile(upload_dataset_path) and not os.path.isdir(upload_dataset_path):
            raise ValueError(UPLOAD_FILE_DIR_ERROR.format(dataset_path=upload_dataset_path))

        is_success, dataset = self.get(dataset_id)
        if not is_success:
            raise ValueError(INVALID_DATASET_ID_ERROR.format(dataset_id=dataset_id))

        access_key = dataset.access_key.access_key
        secret_key = dataset.access_key.secret_key
        minio_service = MinioService(access_key=access_key, secret_key=secret_key)

        if os.path.isdir(upload_dataset_path):
            return minio_service.upload_directory_recursive(
                bucket_name=dataset.bucket.bucket_name,
                source_directory=upload_dataset_path,
                prefix=prefix
            )
        else:
            return minio_service.upload_file(
                bucket_name=dataset.bucket.bucket_name,
                file_path=upload_dataset_path,
                prefix=prefix
            )

    def download_dataset(
        self,
        dataset_id: int,
        local_path: str,
        prefix: Optional[str] = ""
    ):
        is_success, dataset = self.get(dataset_id)
        if not is_success:
            raise ValueError(INVALID_DATASET_ID_ERROR.format(dataset_id=dataset_id))
        access_key = dataset.access_key.access_key
        secret_key = dataset.access_key.secret_key
        try:
            minio_service = MinioService(access_key=access_key, secret_key=secret_key)
            return minio_service.download_directory_recursive(
                bucket_name=dataset.bucket.bucket_name,
                local_path=local_path, prefix=prefix
            )
        except Exception as e:
            return False, str(e)

    def download_load_json_file(
        self, dataset_id, file_path: str
    ):
        is_success, dataset = self.get(dataset_id)
        if not is_success:
            raise ValueError(INVALID_DATASET_ID_ERROR.format(dataset_id=dataset_id))
        access_key = dataset.access_key.access_key
        secret_key = dataset.access_key.secret_key
        try:
            minio_service = MinioService(access_key=access_key, secret_key=secret_key)
            return minio_service.download_load_json_file(
                dataset.bucket.bucket_name,
                file_path
            )
        except Exception as e:
            return False, str(e)

    def list_datasets(self):
        """
        Lists and displays datasets in a formatted table.
        """
        is_success, dataset_response = self.list()

        if not is_success:
            print(DATASET_LISTING_ERROR)
            return

        datasets_table = prepare_datasets_table(dataset_response)
        if datasets_table:
            print(datasets_table)
        else:
            print(NO_DATASETS_AVAILABLE)

    def load_dataset_file(
        self,
        dataset_id: int,
        file_path: str,
        streaming: Optional[bool] = True,
    ):
        """
        Load/Stream a dataset from an EOS bucket using dataset ID and file path.

        Args:
            dataset_id (int): The ID of the dataset to load.
            file_path (str): The file path within the S3 bucket to load.
            streaming (bool): Whether to load the dataset in streaming mode. Defaults to True.

        Returns:
            Union[Dataset, DatasetDict, IterableDataset, IterableDatasetDict]:
            A dataset object depending on the file and whether streaming is enabled.
        """
        is_success, dataset = self.get(dataset_id)
        if not is_success:
            raise ValueError(INVALID_DATASET_ID_ERROR.format(dataset_id=dataset_id))

        file_extension = get_file_extension(file_path)
        if file_extension not in ALLOWED_FILE_FORMATS:
            raise ValueError(ALLOWED_FILES_ERROR.format(file_formats=ALLOWED_FILE_FORMATS))

        try:
            # EOS bucket credentials
            access_key = dataset.access_key.access_key
            secret_key = dataset.access_key.secret_key
            s3_path = f"{dataset.bucket.bucket_url}{file_path}"
            endpoint_url = dataset.bucket.endpoint

            # Load the dataset using the load_dataset
            # This implementation of load_dataset is in experimental stage
            return load_dataset(
                file_extension,
                data_files=s3_path,
                streaming=streaming,
                storage_options={
                    'key': access_key,
                    'secret': secret_key,
                    'client_kwargs': {
                        'endpoint_url': endpoint_url,
                    }
                }
            )
        except Exception as e:
            raise ValueError(DATASET_LOAD_ERROR.format(error_message=str(e)))

    @staticmethod
    def help():
        """
        Provides help information about the Datasets class, its methods, and usage examples.
        """
        help_text = """
    Datasets Class Help
    ===================
    This class provides functionalities to interact with datasets.
    Available methods:
        1. create(name, bucket_name=None, encryption_enable=False, encryption_type=None, bucket_type, description): Creates a new dataset with the provided
        name, bucket name, bucket type, and description. Bucket name is not required if bucket_type='managed'.
        2. get(bucket_name): Retrieves information about a specific dataset using its bucket name.
        3. list(): Retrieves raw dataset information from the API. Returns a response object containing dataset details.
        4. list_datasets(): Fetches dataset information using the `list` method and prints a formatted table of datasets.
        5. delete(dataset_id): Deletes a dataset with the given dataset_id.
        6. upload_dataset(dataset_path, prefix, dataset_id=None): Uploads the dataset to the bucket.
        7. download_dataset(dataset_id, local_path, prefix=""): Downloads the dataset to a local path.
        8. load_dataset_file(dataset_id, file_path, streaming=True): Streams a dataset from an EOS bucket using dataset ID
        and file path. Can optionally stream the dataset.
        9. help(): Displays this help message.

    Example Usages:
        datasets = Datasets(team=123, project=456)
        datasets.create(name='test-dataset', bucket_name='dataset-bucket', bucket_type='managed', description='Test Dataset')
        datasets.get('dataset-bucket')
        datasets.list()
        datasets.list_datasets()
        datasets.delete(dataset_id=236)
        datasets.push_dataset(dataset_path='path/to/dataset', prefix='', dataset_id=None)
        datasets.download_dataset(dataset_id=123, local_path='path/to/local/dir', prefix='')
        datasets.load_dataset_file(dataset_id=123, file_path='path/to/file', streaming=True)
        """
        print(help_text)



# Payload constants
WANDB_USERNAME = "wandB_username"
WANDB_KEY = "wandB_key"
WANDB_PROJECT = "wandB_project_name"
INTEGRATION_DETAILS = "integration_details"
NAME = "name"
HUGGINGFACE_TOKEN = "hugging_face_token"
INTEGRATION_TYPE = "integration_type"

# Basic constants
HUGGING_FACE = "hugging_face"
WEIGHTS_BIASES = "weights_biases"
INTEGRATION_TYPES = [HUGGING_FACE, WEIGHTS_BIASES]

# Error constants
INTEGRATION_TYPE_ERROR = "Integration type '{given_type}' is invalid, it must be in {integration_types}"
WANDB_USERNAME_ERROR = "Please provide a Weights & Biases username."
WANDB_PROJECT_ERROR = "Please provide a Weights & Biases project name."
INTEGRATION_LIST_ERROR = "Unable to list Integrations."
SHOW_DETAILS_ERROR = "Unable to show Integration details."



from typing import Optional

from prettytable import PrettyTable
import requests

from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.distributed_jobs.helpers import (convert_to_base64,
                                                            decode_base64)
from e2enetworks.constants import CLIENT_NOT_READY_MESSAGE
from e2enetworks.cloud.tir.utils import prepare_object


class DistributedJobClient:
    def __init__(self,
                 team: Optional[str] = "",
                 project: Optional[str] = ""
                 ):
        if not client.Default.ready():
            raise ValueError(CLIENT_NOT_READY_MESSAGE)
        if project:
            client.Default.set_project(project)
        if team:
            client.Default.set_team(team)

    def submit_pytorch_job(
        self,
        name: str,
        image: str,
        job_cpu: int,
        job_gpu: int,
        job_memory: int,
        cluster_id: int,
        master_replica: int,
        master_commands: list[str],
        worker_replica: int,
        worker_commands: list[str],
        sfs_id: int = None,
        sfs_mount_path: str = "/mnt/data",
        image_type: str = "public",
        image_pull_policy: str = "IfNotPresent",
        env: list = [],
        _get_raw: bool = False,
        **kwargs
    ):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/?"
        payload = {"name": name,
                   "job_type": "PytorchJob",
                   "job_cpu": job_cpu,
                   "job_gpu": job_gpu,
                   "job_memory": job_memory,
                   "cluster_id": cluster_id,
                   "sfs_id": sfs_id,
                   "sfs_mount_path": sfs_mount_path,
                   "master_replica": master_replica,
                   "master_commands": convert_to_base64(str(master_commands)),
                   "worker_replica": worker_replica,
                   "worker_commands": convert_to_base64(str(worker_commands)),
                   "image_url": image,
                   "image_type": image_type,
                   "image_pull_policy": image_pull_policy,
                   "environmentVariable": env}
        if not sfs_id:
            payload.pop("sfs_id", "")
        req = requests.Request('POST', url, json=payload)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def list_jobs(self,
                  _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def list_clusters(self,
                      _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/cluster/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def describe_cluster(self,
                         cluster_id: int,
                         _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/cluster/{cluster_id}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def list_sfs(self,
                 _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/sfs/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def describe_job(self,
                     job_id: int,
                     _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def delete_job(self,
                   job_id: int,
                   _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/?"
        req = requests.Request('DELETE', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def terminate_job(self,
                      job_id: int,
                      _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/?&action=terminate&"
        req = requests.Request('PUT', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def get_job_pods(self,
                     job_id: int,
                     _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/pods/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def get_job_events(self,
                       job_id: int,
                       _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/events/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def get_job_logs(self,
                     job_id: int,
                     pod_name: str,
                     _get_raw: bool = False):
        url = f"{client.Default.gpu_team_projects_path()}/distributed_jobs/jobs/{job_id}/logs/{pod_name}/?"
        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        return prepare_object(response, _get_raw)

    def show_job_logs(self,
                      job_id: int,
                      pod_name: str):
        is_success, data = self.get_job_logs(job_id, pod_name)
        if is_success:
            print(data.logs)

    def show_job_collective_details(self,
                                    job_id: int):
        is_detail_success, details_resp = self.describe_job(job_id)
        is_pod_success, pods_resp = self.get_job_pods(job_id)
        if is_pod_success == is_detail_success == True:
            main_table = PrettyTable()
            main_table.field_names = ["Field", "value"]
            # Add nested tables to the main table
            for key, value in details_resp.__dict__.items():
                if key not in ['id', 'name', 'job_type', 'status']:
                    continue
                main_table.add_row([key, value])
            main_table.add_row(['', ''])
            nested_table = self._create_pods_sub_table(pods_resp)
            main_table.add_row(['pods', nested_table.get_string()])
            print(main_table)
            print(f"for more info visit : https://gpu-notebooks.e2enetworks.com/teams/{client.Default.team()}/projects/{client.Default.project()}/distributedtraining/?section=training&job_id=1&tab=pod_logs")

    def _create_pods_sub_table(self, pods_array):
        table = PrettyTable()
        table.field_names = ["name", "last_state"]
        for pod in pods_array:
            table.add_row([pod.pod_name, pod.status])
        return table

    @staticmethod
    def help():
        help_text = """
        jobClient Class Help:

        This class provides methods for interacting with job-related operations.
        Before using these methods, make sure to initialize the client using:
        - Using e2enetworks.cloud.tir.init(...)

        Available Methods:
        1. submit_pytorch_job(
            self,
            name: str,
            image: str,
            job_cpu: int,
            job_gpu: int,
            job_memory: int,
            cluster_id: int,
            sfs_id: int,
            sfs_mount_path: str,
            master_replica: int,
            master_commands: list[str],
            worker_replica: int,
            worker_commands: list[str],
            image_type="public",
            image_pull_policy="public",
            environmentVariable: list = [],
            **kwargs
        ):
            - Create a new job.

        2. list_jobs()
            - List existing jobs.

        3. describe_job(job_id)
            - Get details of a specific job.

        4. delete_job(job_id)
            - Delete a specific job.

        5. terminate_job(job_id)
            - Stop a specific job.

        6. list_clusters()
            - List currently available clusters for job.

        7. list_sfs()
            - List available sfs for job.

        Note: Certain methods require specific arguments. Refer to the method signatures for details.
        """
        print(help_text)



from prettytable import PrettyTable


def prepare_datasets_table(dataset_response):
    datasets_table = PrettyTable()
    datasets_table.field_names = ['Dataset ID', 'Dataset Name', 'Encryption Type', 'Created At', 'Updated At', 'Status']
    for item in dataset_response:
        datasets_table.add_row([
            item.id,
            item.name,
            item.encryption_type,
            item.created_at,
            item.updated_at,
            item.status
        ])
    return datasets_table



import requests
from prettytable import PrettyTable

from e2enetworks.cloud.tir import client
from e2enetworks.cloud.tir.helpers import (cpu_plan_short_code,
                                           gpu_plan_short_code)
from e2enetworks.cloud.tir.utils import prepare_object
from e2enetworks.constants import BASE_GPU_URL, INFERENCE, NOTEBOOK, PIPELINE, VECTOR_DB


class Plans:
    def __init__(self):
        client_not_ready = (
            "Client is not ready. Please initiate client by:"
            "\n- Using e2enetworks.cloud.tir.init(...)"
        )
        if not client.Default.ready():
            raise ValueError(client_not_ready)

    def list_endpoint_plans(self,framework=None):
        skus = self.list(INFERENCE,framework=framework)
        
        for sku in skus["CPU"]:
            if sku.get("is_inventory_available") and not sku["is_free"]:
                print(cpu_plan_short_code(sku))
        for sku in skus["GPU"]:
            if sku.get("is_inventory_available") and not sku["is_free"]:
                print(gpu_plan_short_code(sku))
        return skus

    def list(self, service, image_version_id=None, framework=None):

        if type(service) != str:
            print(f"Service - {service} Should be String")
            return
        
        Attributes = [NOTEBOOK, PIPELINE, INFERENCE, VECTOR_DB]
        if service not in Attributes :
            raise ValueError("service must be in %s" % Attributes)

        if service == NOTEBOOK and image_version_id and type(image_version_id) != int:
            print(f"Image ID - {image_version_id} Should be Integer")
            return

        if service == INFERENCE and not framework :
            raise ValueError("framework is required")

        if service == INFERENCE and type(framework) != str:
            print(f"Image ID - {framework} Should be String")
            return

        if service == INFERENCE:
            url = f"{BASE_GPU_URL}gpu_service/sku/?framework={framework}&service={service}&"

        if service == NOTEBOOK:
            image_version_id = image_version_id if image_version_id else ""
            url = f"{BASE_GPU_URL}gpu_service/sku/?image_version_id={image_version_id}&service={service}&"

        if service == PIPELINE :
            url = f"{BASE_GPU_URL}gpu_service/sku/?service={service}&"

        if service == VECTOR_DB:
            url = f"{BASE_GPU_URL}gpu_service/sku/?service={service}&"

        req = requests.Request('GET', url)
        response = client.Default.make_request(req)
        skus = {}
        if response.status_code == 200:
            skus = response.json()["data"]
        return skus

    def get_skus_list(self, service, image_version_id=None, model_id=None, framework=None):
        return self.list(service, image_version_id, framework)

    def get_plans_name(self, service, image_version_id=None, model_id=None, framework=None):
            plans = self.list(service, image_version_id, framework)
            for sku in plans["CPU"]:
                if sku.get("is_inventory_available") and not sku["is_free"]:
                    print(cpu_plan_short_code(sku))
            for sku in plans["GPU"]:
                if sku.get("is_inventory_available") and not sku["is_free"]:
                    print(gpu_plan_short_code(sku))

    def get_plans_list(self, service, image_version_id=None, framework=None):
        plans = self.list(service, image_version_id, framework)
        cpu_skus, gpu_skus = plans["CPU"], plans["GPU"]
        plans_table = PrettyTable()
        plans_table.field_names = ['name', 'series', 'cpu', 'gpu', 'memory',
                                   'sku_item_price_id', 'sku_type', 'committed_days', 'unit_price']
        self.insert_plans_in_table(cpu_skus, plans_table)
        self.insert_plans_in_table(gpu_skus, plans_table)
        return plans, plans_table

    def insert_plans_in_table(self, skus_collection, plans_table):
        for sku in skus_collection:
            for sku_item_price in sku["plans"]:
                if not sku["is_free"]:
                    plans_table.add_row([sku['name'], sku['series'], sku['cpu'], sku['gpu'], sku['memory'],
                                        sku_item_price['sku_item_price_id'], sku_item_price['sku_type'], sku_item_price['committed_days'], sku_item_price['unit_price']])

    @staticmethod
    def help():
        print("Sku Class Help")
        print("\t\t================")
        print("\t\tThis class provides functionalities to interact with Plans.")
        print("\t\tAvailable methods:")
        print("\t\t- list_endpoint_plans: List Available Endpoint Plans")
        print("\t\t1. list(service, image_id): Lists all Plans for given image_id and service.\n")
        print("\t\t Allowed Services List - ['notebook', 'inference_service', 'pipeline']")
        # Example usages
        print("\t\tExample usages:")
        print("\t\tskus = Plans()")
        print("\t\tskus.list('inference')")



from minio import Minio
from minio.error import S3Error
from e2enetworks.constants import S3_ENDPOINT
from e2enetworks.cloud.tir.constants import WRONG_PATH_ERROR
import os
import time
import io
import tempfile
import json


class MinioService:

    def __init__(self, access_key, secret_key, endpoint=S3_ENDPOINT):
        self.endpoint = endpoint
        self.access_key = access_key
        self.secret_key = secret_key
        self.client = Minio(endpoint, access_key, secret_key)

    def upload_directory_recursive(self, bucket_name, source_directory, prefix=""):
        logs = []
        success = True

        for root, dirs, files in os.walk(source_directory):
            for file in files:
                file_path = os.path.join(root, file)
                object_name = os.path.join(prefix, os.path.relpath(file_path, source_directory))

                try:
                    self.client.fput_object(bucket_name, object_name, file_path)
                    log_msg = f"Uploaded {object_name}"
                    logs.append(log_msg)
                except S3Error as e:
                    log_msg = f"Error uploading {object_name}: {e}"
                    logs.append(log_msg)
                    success = False

        return success, logs

    def upload_file(self, bucket_name, file_path, prefix=""):
        object_name = f"{prefix}{os.path.basename(file_path)}" if prefix.endswith('/') else f"{prefix}/{os.path.basename(file_path)}"

        try:
            self.client.fput_object(bucket_name, object_name, file_path)
            return True, f"Uploaded {object_name}"
        except Exception as err:
            return False, f"Error uploading {object_name}: {err}"

    def download_directory_recursive(self, bucket_name, local_path, prefix=""):
        logs = []
        success = True
        objects = self.client.list_objects(bucket_name=bucket_name, prefix=prefix, recursive=True)
        if not objects:
            raise ValueError(WRONG_PATH_ERROR)

        for obj in objects:
            object_name = obj.object_name
            try:
                local_file_path = f"{local_path}/{object_name}"
                self.client.fget_object(bucket_name, object_name, local_file_path)
                logs.append(f"Downloaded: {object_name} to {local_file_path}")
            except Exception as err:
                logs.append(f"Error downloading {object_name}: {err}")
                success = False
        return success, logs

    def update_cache_file(self, bucket_name):
        print("Updating Cache info file")
        object_name = ".tir-cache"
        try:
            current_timestamp = str(int(time.time())).encode()
            current_timestamp_as_a_stream = io.BytesIO(current_timestamp)
            self.client.put_object(bucket_name, object_name, current_timestamp_as_a_stream, len(current_timestamp))
            print("Cache Info Updated")
        except S3Error as err:
            print("Error Occured while updating cache file")
            print(f"Error: {err}")

    def download_load_json_file(
        self, bucket_name: str, object_path: str
    ):
        try:
            with tempfile.NamedTemporaryFile(mode="r+", delete=True, suffix=".json") as temp_file:
                self.client.fget_object(bucket_name, object_path, temp_file.name)
                temp_file.seek(0)
                with open(temp_file.name, "r") as file:
                    content = file.read()
                    json_content = json.loads(content)
                return True, json_content
        except Exception as e:
            return False, str(e)



from copy import deepcopy
import base64
import os
import string
import random
from e2enetworks.constants import BASE_URL_MODEL_API_CLIENT
from e2enetworks.cloud.tir.constants import (
    MODEL_NAME_TO_URL_PATH_MAPPING, MODEL_API_DEFAULT_DATA,
    MODELS_API_DATA_FORMATS, FILE_PATH_EXTENSION_ERROR
)


def get_random_string(N):
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=N))


def cpu_plan_short_code(sku):
    return f"{sku.get('series')}-{sku.get('sku_type').split('-')[0]}-{sku.get('cpu')}-" \
           f"{sku.get('memory')}-{sku.get('gpu') if sku.get('gpu') else 0}"


def gpu_plan_short_code(sku):
    # In split function "-" should be used instead of "."
    return f"{sku.get('series')}-{sku.get('sku_type').split('-')[0]}-{sku.get('cpu')}-" \
           f"{sku.get('memory')}-{sku.get('gpu') if sku.get('gpu') else 0}"


def plan_to_sku_id(skus, plan):
    if plan.startswith("CPU"):
        for sku in skus["CPU"]:
            if cpu_plan_short_code(sku) == plan:
                return sku.get("sku_id")
    else:
        for sku in skus["GPU"]:
            if gpu_plan_short_code(sku) == plan:
                return sku.get("sku_id")
    return False


def plan_name_to_sku_id(skus, sku_unique_name):
    if sku_unique_name.startswith("CPU"):
        for sku in skus["CPU"]:
            if not sku["is_free"] and cpu_plan_short_code(sku) == sku_unique_name:
                return sku.get("sku_id")
    else:
        for sku in skus["GPU"]:
            if not sku["is_free"] and gpu_plan_short_code(sku) == sku_unique_name:
                return sku.get("sku_id")
    return False


def plan_name_to_sku_item_price_id(skus, sku_unique_name):
    if sku_unique_name.startswith("CPU"):
        for sku in skus["CPU"]:
            if not sku["is_free"] and cpu_plan_short_code(sku) == sku_unique_name:
                for plan in sku["plans"]:
                    if plan["committed_days"] == 0:
                        return plan.get("sku_item_price_id")
    else:
        for sku in skus["GPU"]:
            if not sku["is_free"] and gpu_plan_short_code(sku) == sku_unique_name:
                for plan in sku["plans"]:
                    if plan["committed_days"] == 0:
                        return plan.get("sku_item_price_id")
    return False


def get_model_url(model_name, namespace):
    if model_name not in MODEL_NAME_TO_URL_PATH_MAPPING:
        return False, model_name
    url = BASE_URL_MODEL_API_CLIENT + MODEL_NAME_TO_URL_PATH_MAPPING[model_name].format(namespace=namespace)
    return True, url


def get_formated_data_for_model(model_name, parameters):
    extra_keys = []
    fromatted_data = deepcopy(MODEL_API_DEFAULT_DATA)
    model_formatted_data = deepcopy(MODELS_API_DATA_FORMATS).get(model_name)
    for input_key in parameters:
        if input_key in model_formatted_data:
            key_data = model_formatted_data.get(input_key)
            key_data["data"].append(parameters.get(input_key))
            fromatted_data["inputs"].append(key_data)
        else:
            extra_keys.append(input_key)
    if len(extra_keys):
        return False, extra_keys
    return True, fromatted_data


def convert_to_base64(argument):
    json_bytes = argument.encode('utf-8')
    base64_bytes = base64.b64encode(json_bytes).decode('utf-8')
    return base64_bytes


def get_file_extension(file_path):
    _, file_extension = os.path.splitext(file_path)
    file_extension = file_extension.lower().replace('.', '')
    if not file_extension:
        raise ValueError(FILE_PATH_EXTENSION_ERROR.format(file_path=file_path))
    return file_extension
